#include <string.h>
#include <stdlib.h>
#include "spawnio.h"

#ifndef WIN32

#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/wait.h>

static char *readbuf(int fd)
{
	ssize_t ret;
	char buff[4096];
	char *retbuff=NULL;
	int recv = 0;
	while ((ret = read(fd, buff, 4096))!=0)
	{
		char *p = malloc(recv+ret+1);
		if (recv>0)
		{
			memcpy(p, retbuff, recv);
			free(retbuff);
		}
		memcpy(p+recv, buff, ret);
		p[ret+recv]=0;
		recv+=ret;
		retbuff = p;
	}
	if (recv==0)
	{
		retbuff = (char *) malloc(1);
		retbuff[0] = 0;
	}
	return retbuff;
}

char *spawnio(char *cmd[])
{
	int pipes[2];

	if (pipe(pipes)!=0)
	{
		return NULL;
	}

	switch(fork())
	{
		case 0: /*Child*/
			close(pipes[0]);
			close(1);
			close(2);
			dup(pipes[1]);
			dup(pipes[1]);
			execvp(cmd[0], cmd);
			exit(0);
		case -1: return NULL;
		default: /*Parent*/
		{
			char *p;
			int status;

			close(pipes[1]);
			p = readbuf(pipes[0]);
			wait(&status);
			close(pipes[0]);
			return p;
		}

	}
}

#else
#include <windows.h>
#pragma warning(disable : 4996)

static char *readpipe(HANDLE pipe)
{
    DWORD ret;
	char buff[4096];
	char *retbuff=NULL;
	int recv = 0;
        
    while (ReadFile(pipe, buff, 4096, &ret, NULL))
	{
		char *p = malloc(recv+ret+1);
		if (recv>0)
		{
			memcpy(p, retbuff, recv);
			free(retbuff);
		}
		memcpy(p+recv, buff, ret);
		p[ret+recv]=0;
		recv+=ret;
		retbuff = p;
	}
    CloseHandle(pipe);	
    if (recv==0)
    {
	retbuff = (char *) malloc(1);
	retbuff[0] = 0;
     }
     return retbuff;
}

static int pipe(HANDLE *r, HANDLE *w)
{
    SECURITY_ATTRIBUTES sa; 
    sa.nLength = sizeof(SECURITY_ATTRIBUTES); 
    sa.bInheritHandle = TRUE; 
    sa.lpSecurityDescriptor = NULL; 
    if (!CreatePipe(r, w, &sa, 0))
    {
        return -1;
    }
    SetHandleInformation( *r, HANDLE_FLAG_INHERIT, 0);
    return 0;
}

char *spawnio(char *cmd[])
{
    char cmdline[1024];
    char **arg = cmd;
    STARTUPINFO startupinfo;
    PROCESS_INFORMATION procinfo;
    HANDLE stdoutRead, stdoutWrite;
    cmdline[0] = 0;
    strcat(cmdline, *arg);
    while (*(++arg)!=NULL)
    {
        strcat(cmdline, " ");
        strcat(cmdline, *arg);
    }

    if (pipe(&stdoutRead, &stdoutWrite)!=0)
    {
        return NULL;
    }

    ZeroMemory( &startupinfo, sizeof(STARTUPINFO) );
    startupinfo.cb = sizeof(STARTUPINFO);
    startupinfo.dwFlags=STARTF_USESTDHANDLES;
    startupinfo.hStdOutput = stdoutWrite;
    startupinfo.hStdError = stdoutWrite;

    if (CreateProcess(cmd[0], cmdline, 
        NULL, NULL, TRUE, CREATE_NO_WINDOW, NULL, NULL, 
        &startupinfo, &procinfo))
    {
        CloseHandle(procinfo.hThread);
        CloseHandle(procinfo.hProcess);
        CloseHandle(stdoutWrite);
        return readpipe(stdoutRead);
    }
    else
    {
        CloseHandle(stdoutWrite);
        CloseHandle(stdoutRead);
        return NULL;
    }
}

#endif
static void rtrim(char *p)
{
	char *pp = p + strlen(p)-1;
	while (pp>p && *pp==' ')
	{
		*pp = 0;
		pp--;
	}
}




char **spawniolines(char *cmd[])
{
	static char **lines = NULL;
	int i;
	static char *buff;
	int len;
	int numlines = 0;
	char *t;

	/* Free last iteration */
	if (lines!=NULL)
	{
		free(lines);
		free(buff);
		lines = NULL;
	}

	/* Call and return if failed */	
	buff = spawnio(cmd);
	if (buff==NULL)
	{
		return NULL;
	}
	len = (int)strlen(buff);

	/* Count lines */
	for (i=0;i<len;++i)
	{
		if (buff[i]=='\n')
		{
			numlines++;
		}
	}

	lines = malloc(sizeof(char*)*(numlines+2));
	numlines = 0;

	t = strtok(buff, "\n");

	while (t!=NULL)
	{
		rtrim(t);
		lines[numlines++] = t;
		t = strtok(NULL, "\n");
	}
	lines[numlines] = NULL;
	return lines;
}
