/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifndef WIN32


#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>

#include <netdb.h>

#ifdef FreeBSD
#include <sys/socket.h>
#include <netinet/in_systm.h>
#endif

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#include "ciao.h"
#include "onping.h"


static u_short in_cksum( u_short *addr, int len);

#define PACKETDATA "A_GOOD_LUCK_PACKET"

static u_short seq=0;

static int opensocket();

static int pinger(int s, struct sockaddr_in *to, int ident)
{
	u_char packethdr[IP_MAXPACKET];
	u_char *packet = packethdr + sizeof(struct ip);
	int datalen = strlen(PACKETDATA);
	u_char *datap = packet + ICMP_MINLEN;
	struct icmp *icp = (struct icmp *)packet;
	int cc;
	int ret;

	/* Fill Data */
	memcpy(datap, PACKETDATA, datalen);

	/*icmp packet */
	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_seq = htons(seq++);
	icp->icmp_id = ident;
	icp->icmp_cksum = 0;

	cc = ICMP_MINLEN + datalen;

	icp->icmp_cksum = in_cksum((u_short*)icp, cc);

	ret = sendto(s, (char*)packet, cc, 0, 
		(struct sockaddr*)to, sizeof(struct sockaddr_in));
	if (ret!=cc)
	{
		return PING_FAILED;	
	}
	return 0;
}

int validatepacket(u_char *packet, int recvlen, int ident)
{
	struct ip *ip = (struct ip *) packet;
	int hlen = ip->ip_hl << 2;
	struct icmp *icp = (struct icmp *) (packet + hlen);
	int expectedlen = strlen(PACKETDATA);
	char *data = icp->icmp_data;

	/*int cc;
	for (cc=0;cc<recvlen;++cc)
	{
		if (packet[cc]>' ' ) 
		{
			printf("%c", packet[cc]);
		}
		else
		{
			printf("[%d]", packet[cc]);
		}
	}
	printf("\n");*/
	

	if (recvlen < hlen + ICMP_MINLEN) 
	{
		return ECHO_CORRUPT;
	}

	if (icp->icmp_id!=ident)
	{
		return 1; /* Not our packet */
	}

	if ( data+expectedlen-(char*)packet>recvlen)
	{
		return ECHO_CORRUPT;
	}
 	
	if (memcmp(data, PACKETDATA, expectedlen)!=0)
	{
		return ECHO_CORRUPT;
	}

	if (ntohs(icp->icmp_seq)!=seq-1)
	{
		return 1; /* Return from a previous request */
	}

	return 0;
}

int recvecho(int s, int ident)
{
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_in from;
	int ret;
	u_char packethdr[IP_MAXPACKET];
	u_char *packet = packethdr + sizeof(struct ip);

	bzero(&msg, sizeof(msg));
	msg.msg_name = (caddr_t)&from;
	msg.msg_namelen = sizeof(from);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	iov.iov_base = packet;
	iov.iov_len = IP_MAXPACKET;

	if ((ret = recvmsg(s, &msg, 0))<0)
	{
		return RESOURCE_ERROR;
	}
	return validatepacket(packet, ret, ident);/* 1 = not mine */
}

int waitforecho(int s, int milisec)
{
	fd_set rfds;
	struct timeval timeout;
	int ret;

	
	FD_ZERO(&rfds);
	FD_SET(s, &rfds);
	timeout.tv_sec = milisec/1000;;
	timeout.tv_usec = (milisec%1000)*1000;

	ret = select(s+1, &rfds, NULL, NULL, &timeout);

	switch(ret)
	{
		case -1: return RESOURCE_ERROR;
		case 0: return PING_FAILED;
		default: return 0;
	}
}

int adjustErr(int code, char *errmsg, int first_timeout, int final_timeout)
{
	switch(code)
	{
		case OKAY:
			sprintf(errmsg, "Ping reestablished within timeout");
			break;
		case HOST_RESOLVE_ERROR:
			sprintf(errmsg, "Can't translate host/ip");
			break;
		case ECHO_CORRUPT:
			sprintf(errmsg, "Return packet received but it was corrupted");
			break;
		case PING_FAILED: /* Here it still means timeout */
			sprintf(errmsg, "Ping timed out after %d seconds", final_timeout/1000);
			break;
		case QTIMEOUT: /* The first_timeout failed but the final_timeout was good */
			sprintf(errmsg, "Response time exceeded %dms timeout", first_timeout);
			break;
		case RESOURCE_ERROR:
		default:
			sprintf(errmsg, "Ping failed: %s", strerror(errno));
			code = PING_FAILED;
			break;
	}
	return code;
}

int recvmyecho(int sock, int ident, int timeout)
{
        time_t start = time(NULL);
	int ret;
	do { /* The loop is here because I can get packets from other instances of ping */
		if ((ret = waitforecho(sock, timeout))==0)
		{
			ret = recvecho(sock, ident);
		}
	} while (ret==1 && time(NULL)<=start+timeout/1000+1);
	if (ret==1)
	{
		ret = PING_FAILED;
	}
	return ret;
}

int ping(char *host, int first_timeout, int final_timeout)
{
	struct sockaddr_in whereto;  
	struct sockaddr_in *to;
	int ret;
	int ident = getpid() & 0xFFFF;
	int sock;

	/* Resolve Address */
	bzero(&whereto, sizeof(whereto));
	to = &whereto;
	to->sin_family = AF_INET;
	/*to->sin_len = sizeof(*to);*/
	if (inet_aton(host, &to->sin_addr) == 0)
	{
		struct hostent *hp = gethostbyname2(host, AF_INET);
		if (!hp)
		{
			return HOST_RESOLVE_ERROR;
		}
		if ((unsigned)hp->h_length > sizeof(to->sin_addr))
		{
			return HOST_RESOLVE_ERROR;
		}
		memcpy(&to->sin_addr, hp->h_addr_list[0], sizeof to->sin_addr);
	}

	sock = opensocket();
	if (sock<0)
	{
		return RESOURCE_ERROR;
	}

        ret = pinger(sock, to, ident);
	
        if (ret==0) 
        {
		if (final_timeout>first_timeout)
		{
			ret = recvmyecho(sock, ident, first_timeout);
			if (ret==PING_FAILED)
			{
				ret = recvmyecho(sock, ident, final_timeout-first_timeout);
				if (ret==OKAY)
				{
					ret = QTIMEOUT;
				}
			}
		}
		else
		{
			ret = recvmyecho(sock, ident, final_timeout);
		}
        }
	close(sock);
        return ret;
}



static u_short in_cksum( u_short *addr, int len)
{
	int nleft, sum;
	u_short *w;
	union {
		u_short us;
		u_char  uc[2];
	} last;
	u_short answer;

	nleft = len;
	sum = 0;
	w = addr;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		last.uc[0] = *(u_char *)w;
		last.uc[1] = 0;
		sum += last.us;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);     /* add hi 16 to low 16 */
	sum += (sum >> 16);		     /* add carry */
	answer = ~sum;			  /* truncate to 16 bits */
	return(answer);
}

static int opensocket()
{
	int hold = IP_MAXPACKET + 128;
	int s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (s==-1)
	{
		iao_logsysmsg("Can't open SOCK_RAW (not admin?)");
		return RESOURCE_ERROR;
	}
	setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&hold, sizeof(hold));
	setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char *)&hold, sizeof(hold));
	return s;
}


static int pingretry(char *host, int first_timeout, int final_timeout)
{
	int retries = 3;
	int ret;

	while (retries>0)
	{
		ret = ping(host, first_timeout, final_timeout);
		switch(ret)
		{
		case ECHO_CORRUPT:
		case PING_FAILED:
			retries--;
			sleep(3);
			break;
		case OKAY:
		case HOST_RESOLVE_ERROR:
		case RESOURCE_ERROR:
		default:
			retries=0;
			break;
		}
	}
	return ret;
}

static int checkhost(hoststate *host, int first_timeout, int final_timeout)
{
	int ret = pingretry(host->host, first_timeout, final_timeout);
	host->actual = adjustErr(ret, host->msg, first_timeout, final_timeout);
	return host->actual==host->reported?0:1;
}

int checkping(hoststate *hosts, int size,int first_timeout, int final_timeout)
{
	int notif = 0;
	int i;

	for (i=0;i<size;++i)
	{
		if (checkhost(hosts+i, first_timeout, final_timeout)==1)
		{
			notif = 1;
		}
	}
	return notif;
}

int pinginit()
{
	int sock = opensocket();
	if (sock<0)
	{
		return -1;
	}
	else
	{
		close(sock);
		return 0;
	}
}

#endif /* !WIN32 */
