/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "ciao.h"

#ifdef WIN32
#pragma warning(disable:4996)
#define strcasecmp stricmp
#define strncasecmp strnicmp
#endif

#define MAX_PROCS 10000

static char **plist = NULL;
static int pnum = 0;
static int pnum_noservices;
static int maxproclen = 0; /* no limit - expect full match */
static void freelist()
{
	if (plist!=NULL)
	{
		char **p = plist;
		while (*p!=NULL)
		{
			if (*p!=NULL)
			{
				free(*p);
			}
			p++;
		}
		free(plist);
		plist = NULL;
	}
}

static void procnum(int num)
{
	int size = sizeof(char*)*(num+1);
	pnum = 0;
	plist = malloc(size);
	memset(plist, 0, size);
}

void noexe(char *s)
{
	char *p = s + strlen(s) - 4;

	if (p>s)
	{
		if (strcasecmp(p,".exe")==0)
		{
			*p = 0;
		}
	}
}

static void addproc(char *name)
{
	plist[pnum] = malloc(strlen(name)+1);
	strcpy(plist[pnum], name);
	#ifdef WIN32 
	noexe(plist[pnum]);
	#endif
    if (pnum<MAX_PROCS-1) 
    {
	    pnum++;
    }
}

int strlennoexe(char *p)
{
	int len = (int)strlen(p);
	char *ext = p-4+len;
	if (ext>p)
	{
		if (strcasecmp(ext, ".exe")==0)
		{
			return len-4;
		}
	}
	return len;
}

int strcasecmpnoexe(char *p1, char *p2)
{
	int len1 = strlennoexe(p1);
	int len2 = strlennoexe(p2);
	if (len1!=len2)
	{
		return -1;
	}
	return strncasecmp(p1, p2, len1);
}

int match(char *p1, char *p2)
{
	if (maxproclen==0)
	{
#ifdef WIN32
		return strcasecmpnoexe(p1,p2)==0;
#else
		return strcasecmp(p1,p2)==0;
#endif
	}
	else
	{
		return strncasecmp(p1, p2, maxproclen)==0;
	}
}

#ifdef WIN32
#include <windows.h>
#include <Tlhelp32.h>
#include <psapi.h>

#define SERVICES_BUFF 4096

int isprocessname(char *s)
{
    int i;
    for (i=0;i<pnum_noservices;++i)
    {
        if (strcasecmp(s, plist[i])==0)
        {
            return 1;
        }
    }
    return 0;
        /*
        This old code doesn't work on 64 bit (when i am 32)
        the idea was to check for each service it happens to be identical to its .exe name
        HANDLE proc = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ, FALSE, pid);
		if (proc!=NULL)
		{
			char exe[1024];
			if (GetModuleBaseName(proc, NULL, exe, 1024)!=0)
			{
				if (strcasecmpnoexe(ss->lpServiceName, exe)!=0)
				{
					addproc(ss->lpServiceName);
				}
			}
			else
			{
				iao_logsysmsg("service: GetModuleBaseName failed");
			}
			CloseHandle(proc);
		}
		else
		{
			char buff[256];
			sprintf(buff, "service: Failed openning proc handle for pid: %d", pid);
			iao_logsysmsg(buff);
		}*/

}


void service(ENUM_SERVICE_STATUS_PROCESS *ss)
{
	DWORD pid = ss->ServiceStatusProcess.dwProcessId;
	if (ss->ServiceStatusProcess.dwCurrentState==SERVICE_RUNNING && pid!=0) 
	{
        // To avoid duplicates
        if (!isprocessname(ss->lpServiceName)) 
        {
            addproc(ss->lpServiceName);
        }
	}
}

static int services()
{
	int ret = -1;
	SC_HANDLE sman = OpenSCManager(NULL, SERVICES_ACTIVE_DATABASE, GENERIC_READ);
    pnum_noservices = pnum;
	if (sman!=NULL)
	{
		int cont;
		DWORD resh = 0;

		do {
			DWORD bytesneeded;
			DWORD servicecount;
			DWORD i = 0;
			char buff[SERVICES_BUFF];
			cont = 0;
			if (EnumServicesStatusEx(sman, SC_ENUM_PROCESS_INFO, SERVICE_WIN32, SERVICE_ACTIVE, 
				(LPBYTE) buff, SERVICES_BUFF, &bytesneeded, &servicecount, &resh, NULL)==0)
			{
				if (GetLastError()!=ERROR_MORE_DATA)
				{
					iao_logsysmsg("services: EnumServiceStatus failed");
					break;
				}
				cont = 1;
			}
			ret = 0;
			while (i<servicecount&&pnum<MAX_PROCS)
			{
				ENUM_SERVICE_STATUS_PROCESS  *ss = (ENUM_SERVICE_STATUS_PROCESS *)buff;
				service(ss+i);
				++i;
			}
		} while (cont);
		CloseServiceHandle(sman);
	}
	else
	{
		iao_logsysmsg("services: Couldn't open services database");
	}
	return ret;
}


static int sysproclist()
{
	HANDLE snap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
	if (snap!=INVALID_HANDLE_VALUE)
	{
		PROCESSENTRY32 pe;
		int cont;
		
		pe.dwSize = sizeof(PROCESSENTRY32);
		cont = Process32First(snap, &pe);
		procnum(MAX_PROCS);
		while (cont==TRUE && pnum<MAX_PROCS)
		{
			addproc(pe.szExeFile);
			cont = Process32Next(snap, &pe);
		}
		CloseHandle(snap);
		services();
		return 0;
	}
	else
	{
		iao_logsysmsg("sysproclist: Can't open process list");
		return -1;
	}	
}

#endif

#ifdef old_Darwin

#include <fcntl.h>
#include <kvm.h>
#include <sys/sysctl.h>

static int sysproclist()
{
	int ret = -1;
	kvm_t *k = kvm_open(NULL, NULL, NULL, O_RDONLY, "sysproclist");
	if (k!=NULL)
	{
		int c;
		struct kinfo_proc *procs=kvm_getprocs(k, KERN_PROC_ALL, 0, &c);
		if (procs!=NULL)
		{
			int i;
			procnum(c);
			for (i=0;i<c;++i)
			{
				addproc(procs[i].kp_proc.p_comm);
			}
			ret = 0;
		}
		else
		{
			iao_logsysmsg("sysproclist: Can't open process list");
		}
		kvm_close(k);
	}
	else
	{
		iao_logmsg("sysproclist: can't open process list (permission?)");
	}
 	return ret;
}
#endif /* old_Darwin */

#ifdef FreeBSD
#define Darwin
#endif
#ifdef Darwin

#include "spawnio.h"

static int sysproclist()
{
	static char *cmd[] = { "/bin/ps","-A","-o", "ucomm", NULL };
        char **p = spawniolines(cmd);
	char **c = p;
	int count = -1;
	int first1 = 1;

	if (p==NULL)
	{
		iao_logsysmsg("sysproclist: spawniolines failed");
		return -1;
	}
	
	while (*c!=NULL)
	{
		count++;
		c++;
	}

	procnum(count);
	while (*p!=NULL)
	{
		if (first1) /* first line is a header */
		{
			first1 = 0;
		}
		else
		{
			addproc(*p);
		}
		p++;
	}
	return 0;
}
#endif /*Darwin */

#ifdef Linux
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#define MAX_PROCS 10000

static int sysproclist()
{
	int ret = -1;
	DIR *dir = opendir("/proc");
	if (dir!=NULL)
	{
		struct dirent *entry;
		procnum(MAX_PROCS);
		ret = 0;
		while ( (entry=readdir(dir))!=NULL && pnum<MAX_PROCS)
		{
			if (isdigit(entry->d_name[0]))
			{
				int fd;
				char buf[1024];
				snprintf(buf, sizeof(buf), "/proc/%s/stat", entry->d_name);
				if ((fd = open(buf, O_RDONLY))>=0)
				{
					int len = read(fd, buf, sizeof(buf));
					close(fd);
					if (len>0)
					{
						char *f;
						char *t;
						buf[len]=0;
						f = strchr(buf, '(');
						t = strchr(f, ')');
						memmove(buf, f+1, t-f-1);
						buf[t-f-1]=0;
						addproc(buf);
					}
				}
			}
		}
		closedir(dir);
	}
	else
	{
		iao_logsysmsg("sysproclist: Can't open process list");
	}
	return ret;
}
#endif /* Linux */

#ifdef SunOS

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#define _STRUCTURED_PROC 1
#include <sys/procfs.h>

static int sysproclist()
{
	int ret = -1;
	maxproclen = PRFNSZ;
	DIR *dir = opendir("/proc");
	if (dir!=NULL)
	{
		struct dirent *entry = readdir(dir);
                procnum(MAX_PROCS);
		while (entry!=NULL && pnum<MAX_PROCS)
		{
			if (entry->d_name[0] != '.')
			{
				int fd;
				char buf[30];
				snprintf(buf, sizeof(buf), "/proc/%s/psinfo", entry->d_name);
				if ((fd = open(buf, O_RDONLY))>=0)
				{
					psinfo_t info;
					if (read(fd, &info, sizeof(psinfo_t)) == sizeof(psinfo_t))
					{
						addproc(info.pr_fname);
					}
					close(fd);
				}
			}
			entry = readdir(dir);
		}
		ret = 0;
		closedir(dir);
	}
	else
	{
		iao_logsysmsg("sysproclist: Can't open process list");
	}
	return ret;
}
#endif /* SunOS */


char **proclist()
{
	freelist();
	if (sysproclist()==0)
	{
		return plist;
	}
	else
	{
		return NULL;
	}
}
