#include "dir.h"

static int dotdot(char *f)
{
        return f[0]=='.' && (f[1]==0 || (f[1]=='.' && f[2]==0));
}
#ifdef WIN32
#include <windows.h>
#include <string.h>
#include <stdio.h>
typedef __int64 int64;
#pragma warning(disable : 4996)

typedef struct {
    HANDLE h;
	WIN32_FIND_DATA fdata;
} dirhandlewin;


dirhandle opendirhandle(char *path)
{
    dirhandlewin *dhw = (dirhandlewin *)malloc(sizeof(dirhandlewin));
    char correctpath[1024];

    if (strchr(path, '*')==NULL) {
        sprintf(correctpath, "%s%s", path, path[strlen(path)-1]=='\\'?"*":"\\*");
        path = correctpath;
    }

    dhw->h = FindFirstFile(path, &(dhw->fdata));
    if (dhw->h==INVALID_HANDLE_VALUE)
    {
        free(dhw);
        dhw = NULL;
    }
    return dhw;
}

int nextfile(dirhandle d, fileinfo *fi)
{
    dirhandlewin *dh = (dirhandlewin*) d;
    int found = 0;
    do {
        if (dh->fdata.dwFileAttributes==INVALID_FILE_ATTRIBUTES && !FindNextFile(dh->h, &(dh->fdata)))
        {
            return 0;
        }

        fi->isdir = dh->fdata.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY;
        if (!(fi->isdir && dotdot(dh->fdata.cFileName)))
        {
            strncpy(fi->name, dh->fdata.cFileName, MAX_FILENAME_LEN);
            fi->name[MAX_FILENAME_LEN-1] = 0;
            fi->size =  ((int64)dh->fdata.nFileSizeHigh * (((int64)MAXDWORD)+1L)) + (int64)dh->fdata.nFileSizeLow;
            found = 1;
        }
        dh->fdata.dwFileAttributes = INVALID_FILE_ATTRIBUTES;
        
    } while (!found);

    return 1;
}


void closedirhandle(dirhandle d)
{
    dirhandlewin *dh = (dirhandlewin*) d;
    FindClose(dh->h);
    free(d);
}

int getfileinfo(char *path, fileinfo *fi)
{
    int result = 0;
    DWORD attr = GetFileAttributes(path);
    strncpy(fi->name, path, MAX_FILENAME_LEN);
    fi->name[MAX_FILENAME_LEN-1] = 0;
    if (attr!=INVALID_FILE_ATTRIBUTES)
    {
        if (attr&FILE_ATTRIBUTE_DIRECTORY)
        {
            fi->isdir = 1;
            fi->size = 0;
            result = 1;
        }
        else
        {
            LARGE_INTEGER largeint;
            HANDLE file = CreateFile( path, GENERIC_READ, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
            if (file!=INVALID_HANDLE_VALUE)
            {
                if (GetFileSizeEx(file, &largeint))
                {
                    fi->size = largeint.QuadPart;
                    fi->isdir = 0;
                    result = 1;
                }
                CloseHandle(file);
            }
        }
    }
    return result;
}

char *currentdirectory()
{	
    static char buff[256];

    GetCurrentDirectory(256, buff);
    return buff;
}


#else
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>


typedef struct {
	DIR *d;
	char path[MAX_FILENAME_LEN];
	int pathlen;
} dirhandleux;


dirhandle opendirhandle(char *path)
{
	DIR *p = opendir(path);
	if (p) 
	{
		dirhandleux *dh = (dirhandleux*)malloc(sizeof(dirhandleux));
		dh->d = p;
		strncpy(dh->path, path, MAX_FILENAME_LEN);
		dh->path[MAX_FILENAME_LEN-1] = 0;
		dh->pathlen = strlen(dh->path);
		return dh;
	}
	else
	{
		return NULL;
	}
}

int nextfile(dirhandle d, fileinfo *fi)
{
	dirhandleux *dh = (dirhandleux*) d;
	struct dirent *p;

	while ((p=readdir(dh->d))) 
	{
		if (p->d_type==DT_DIR && !dotdot(p->d_name)) {
			fi->isdir = 1;	
			fi->size = 0;
			strncpy(fi->name, p->d_name, MAX_FILENAME_LEN);
			fi->name[MAX_FILENAME_LEN-1] = 0;
			return 1;
		} else if (p->d_type==DT_REG && dh->pathlen+strlen(fi->name+2)<MAX_FILENAME_LEN) {
			int ret;
			sprintf(fi->name, "%s/%s", dh->path, p->d_name);
			ret = getfileinfo(fi->name, fi);
			strcpy(fi->name, p->d_name);
			return ret;
		}
	}
	return 0;
}

void closedirhandle(dirhandle d)
{
        dirhandleux *dh = (dirhandleux*) d;
	closedir(dh->d);
	free(dh);
}

int getfileinfo(char *path, fileinfo *fileinfo)
{
	struct stat s;
	if (stat(path, &s)==0)
	{
		fileinfo->isdir = s.st_mode&S_IFDIR;
		fileinfo->size = fileinfo->isdir?0:s.st_size;
		strncpy(fileinfo->name, path, MAX_FILENAME_LEN);
		fileinfo->name[MAX_FILENAME_LEN-1] = 0;
		return 1;
	}
	else
	{
		return 0;
	}
}

char *currentdirectory()
{	
	static char *cd = NULL;
	if (cd==NULL) cd = getcwd(NULL, 0);
	return cd;
}
#endif

#ifdef UNITTEST

#include <stdio.h>
int main(int argc, char *argv[])
{
	char *path  = argc>1?argv[1]:currentdirectory();
    fileinfo fi;
    
    if (getfileinfo(path, &fi))
    {
        if (fi.isdir)
        {
	        dirhandle dh = opendirhandle(path);
	        if (dh==NULL)
	        {
		        printf("Can't open directory: %s\n", path);
	        }
	        else
	        {
		        fileinfo fi;
		        printf("%s:\n", path);
		        while (nextfile(dh, &fi))
		        {
			        printf("%s%s\t\t%lld\n", fi.name, fi.isdir?"/":"", fi.size);
		        }
		        closedirhandle(dh);
	        }
        }
        else
        {
            printf("%s: %d\n", path, fi.size);
        }
    }
    else
    {
        printf("Can't get file info for %s\n", path);
    }

	return 0;
}
#endif
